#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Copyright (c) 2012, Nathan Sheldon.
# http://www.nathansheldon.com/files/NOAA-Weather-Plus.php
#
#	Version 1.1.19
#
#	History:	1.1.19 (09-Jan-2017)
#				* Fixed weather condition icon image names that appeared as
#				  "DualImage" for the icon name due to changes in NOAA data feed.
#				--
#				1.1.18 (25-Jul-2015)
#				* Fixed a bug that caused the weather condition icon name state
#				  to show up as a partial URL instead of the icon filename.
#				--
#				1.1.17 (18-Nov-2014)
#				* Fixed a bug that would cause the plugin to crash if the wind
#				  speed element didn't exist in the weather data.
#				--
#				1.1.16 (27-Nov-2013)
#				* Fixed a bug that would cause the plugin to crash if a current
#				  conditions observation time of "NA" was specified in the XML
#				  downloaded from NOAA.
#				--
#				1.1.15 (05-Oct-2013)
#				* Fixed a bug that could cause the plugin to crash if the NOAA server
#				  connection timed out after a connection was already established.
#				--
#				1.1.14 (30-Jul-2013)
#				* Added the indigoPluginUpdateChecker module (code by Travis CooK)
#				  to facilitate automatic plugin version checking.
#				--
#				1.1.13 (27-May-2013)
#				* Fixed a bug in the update method that could cause the plugin to
#				  crash if there was a problem communicating with NOAA servers.
#				* Updated forecast URL accessed.  The forecast XML feed now appears to
#				  requires that a language be specified in the forecast request.
#				--
#				1.1.12 (22-Jan-2013)
#				* Added more debugging to and updated wind chill calculation algorithm.
#				  It now uses the alternate weather information for calculations if needed.
#				* Made the alternate data gathering and parsing more efficient.
#				--
#				1.1.11 (21-Jan-2013)
#				* Fixed some bugs that would cause the runConcurrentThread to crash
#				  if various forecast data was unavailable.
#				* Added more robust error checking and handling throughout.
#				* Fixed a bug in the Wind Chill calculation formula that caused large
#				  negative numbers.
#				--
#				1.1.10 (20-Dec-2012)
#				* Added more error checking in current conditions section to fix
#				  some rare bugs that would cause the runConcurrentThread to crash
#				  every once in a while.
#				--
#				1.1.9 (29-Nov-2012)
#				* Fixed bug that would cause a plugin crash if the wind degrees data
#				  from NOAA was a string rather than a number.
#				* Improved data parsing for NOAA weather stations located off-shore.
#				* Added waterTemperatureF and waterTemperatureC device states for
#				  off-shore statsions which report that information.
#				--
#				1.1.8 (08-Nov-2012)
#				* Updated dialogs for Indigo 6 UI.
#				* Fixed a bug that could cause the plugin to crash if the wind
#				  direction was blank.
#				* Fixed a bug that could cause the plugin to crash if GPS coordinates
#				  were used for which NOAA has no data and for which NOAA returns
#				  unusable XML code.
#				--
#				1.1.7 (30-Sep-2012)
#				* Improved error message clarity in Indigo log when unable
#				  reach the NOAA server due to DNS or network connectivity issues.
#				--
#				1.1.6 (28-Sep-2012)
#				* Fixed a bug that would cause the runConcurrentThread to
#				  exit unexpectedly if the URL for the data was unreachable.
#				--
#				1.1.5 (17-Sep-2012)
#				* Fixed a bug that would cause a runConcurrentThread crash
#				  while attempting to obtain the moreWeatherInformation
#				  attribute if that attribute didn't exist in the data feed.
#				--
#				1.1.4 (10-Sep-2012)
#				* Fixed a bug that caused the observation time in the primary
#				  data source to always appear to be blank, even if it wasn't.
#				  (Thanks to Tim Hurley for pointing this out).
#				--
#				1.1.3 (1-Sep-2012)
#				* Fixed a bug that would cause the update method to error out
#				  if the observation time was missing an element.
#				--
#				1.1.2 (30-Aug-2012)
#				* Fixed a bug that caused updating to partially fail if there
#				  were more than 3 hazardous weather advisotires for any NOAA
#				  Weather Plus device.
#				--
#				1.1.1 (27-Aug-2012)
#				* Fixed bug that would cause an individual device's thread to
#				  crash if no current observation (or other data) was returned
#				  from the primary NOAA data source.
#				--
#				1.1 (23-Aug-2012)
#				* Added 9 new device states for up to 3 weather advisories
#				  (weatherAdvisory1Text, weatherAdvisory1StartTime, and
#				  weatherAdvisory1EndTime, with the ...2... and ...3... versions).
#				* Changed almost all forecast state names both within the device
#				  definition and in the user interface to make it more clear what
#				  each state represents. Added code to "align" the high and low
#				  temperature forecasts with the bi-daily weather conditions.
#				* Rearranged the order of some state values in the user interface.
#				* Added code to attempt to find and download weather data from
#				  an alternate NOAA weather URL if the information doesn't exist
#				  in the primary XML data obtained.
#				* Changed what is inserted into states when data is unavailable.
#				* Updated some error messages to indicate possible workarounds for
#				  parsing errors.
#				--
#				1.0.5 (09-Aug-2012)
#				* Added code to allow for inconsistent data in relative humidity
#				  and wind speed (Knots) elements of data feed. Changed some error
#				  logging to non-error log entries.
#				--
#				1.0.4 (08-Aug-2012)
#				* Changed error handling for instances when the XML from NOAA
#				  contains no current observation time so that we continue
#				  to parse the current weather but simply warn the user that
#				  it may be out of date, rather than aborting data processing.
#				--
#				1.0.3 (02-Jul-2012)
#				* It seems NOAA has started putting their icon links into a
#				  "medium" subfolder, breaking the icon name filter. This
#				  version fixes that issue.
#				* Added code to enable debugging immediately after closing
#				  the plugin settings window.
#				--
#				1.0.2 (23-Jun-2012)
#				* Corrected error with automatically filling in coordinates.
#				--
#				1.0.1 (22-Jun-2012)
#				* Now automatically fills in latitude and longitude
#				  values in new device setup from Indigo prefs.
#				--
#				1.0 (22-Jun-2012)
#				* Initial release.
#
# Code loosely based on "NOAA Weather" plugin by Perceptive Automation, LLC.
# http://www.perceptiveautomation.com
#

################################################################################
# Imports
################################################################################
import re
import urllib2
from xml.dom.minidom import parseString
import indigoPluginUpdateChecker

################################################################################
# Globals
################################################################################
theUrlBase = u"http://forecast.weather.gov/MapClick.php?FcstType=dwml&lg=english&"
weatherStationListUrl = u"http://w1.weather.gov/xml/current_obs/index.xml"
alternateDataFields = {"currentCondition":'weather',"currentConditionIcon":'icon_url_name',"visibility":'visibility_mi',"temperatureF":'temp_f',"temperatureC":'temp_c',"waterTemperatureF":'water_temp_f',"waterTemperatureC":'water_temp_c',"temperatureString":'temperature_string',"humidity":'relative_humidity',"dewPointF":'dewpoint_f',"dewPointC":'dewpoint_c',"dewPointString":'dewpoint_string',"heatIndexF":'heat_index_f',"heatIndexC":'heat_index_c',"heatIndexString":'heat_index_string',"windDirection":'wind_dir',"windDegrees":'wind_degrees',"windMPH":'wind_mph',"windKnots":'wind_kt',"windString":'wind_string',"pressureInches":'pressure_in',"pressureMillibars":'pressure_mb',"location":'location',"latitude":'latitude',"longitude":'longitude'}

currentConditions = ("currentCondition","currentConditionIcon","temperatureF","temperatureC","humidity","dewPointF","dewPointC","heatIndexF","heatIndexC","waterTemperatureF","waterTemperatureC","visibility","windDirection","windDegrees","windKnots","windMPH","windGustKnots","windGustMPH","windChillF","windChillC","pressureInches","pressureMillibars","location","latitude","longitude","elevation")

forecastLocationData = ("forecastLocation","forecastLatitude","forecastLongitude","forecastElevation")

maxTempVals = ("maxTempToday","maxTempTomorrow","maxTemp2ndDay","maxTemp3rdDay","maxTemp4thDay","maxTemp5thDay","maxTemp6thDay","maxTemp7thDay")

minTempVals = ("minTempTonight","minTempTomorrowNight","minTemp2ndNight","minTemp3rdNight","minTemp4thNight","minTemp5thNight","minTemp6thNight")

forecastPeriods = ("Today", "Tonight", "Tomorrow", "TomorrowNight", "2ndDay", "2ndNight", "3rdDay", "3rdNight", "4thDay", "4thNight", "5thDay", "5thNight", "6thDay", "6thNight", "7thDay")

################################################################################
class Plugin(indigo.PluginBase):
	########################################
	# Class properties
	########################################
	
	########################################
	def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs): 
		indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
		self.debug = pluginPrefs.get("showDebugInfo", False)
		self.debugLog("Initializing Plugin.")
		self.deviceList = []
		self.alternateWeatherXml = ""
		self.alternateWeatherMapXml = ""
		# Load the update checker module.
		self.updater = indigoPluginUpdateChecker.updateChecker(self, 'http://www.nathansheldon.com/files/PluginVersions/NOAA-Weather-Plus.html')
	
	########################################
	def __del__(self):
		self.debugLog("Stopping plugin.")
		indigo.PluginBase.__del__(self)
	
	########################################
	def goToInfoURL(self, valuesDict, typeId, devId):
		self.browserOpen("http://www.perceptiveautomation.com/indigo/location_lookup.html")
		
	# Device Start
	########################################
	def startup(self):
		# Perform an initial version check.
		self.updater.checkVersionPoll()
		
	########################################
	def deviceStartComm(self, device):
		self.debugLog("Starting device: " + device.name)
		# Make sure the device has all the states. (People upgrading
		#   to this version will not have all the states in their
		#   NOAA Weather Plus devices). Using "forecastTodayText"
		#   as the state to look for to indicate an old device.
		# if not device.states.get('forecastTodayText', False):
			# indigo.server.log(u"WARNING: This is a version 1.0.x device. It will be updated to the current \n    version (" + self.pluginVersion + ").  This will cause most forecast state IDs to change and thus all Device State Change triggers \n    based on those states to stop functioning until they've been edited to use the new forecast states.", device.name)
			# This will forece Indigo to update the device with the new states.
			# device.stateListOrDisplayStateIdChanged()
			
		# Update pre-1.1.9 devices to include the water temperature states.
		if not device.states.get('waterTemperatureF', False):
			# This will forece Indigo to update the device with the new states.
			device.stateListOrDisplayStateIdChanged()
			
		# Update the device list and the device states.
		if device.id not in self.deviceList:
			self.deviceList.append(device.id)
			self.update(device)
			
	########################################
	def deviceStopComm(self, device):
		self.debugLog("Stopping device: " + device.name)
		if device.id in self.deviceList:
			self.deviceList.remove(device.id)
	
	########################################
	def closedPrefsConfigUi(self, valuesDict, userCancelled):
		if not userCancelled:
			self.debug = valuesDict.get("showDebugInfo", False)
			if self.debug:
				indigo.server.log("Debug logging enabled")
			else:
				indigo.server.log("Debug logging disabled")
	
	########################################
	def runConcurrentThread(self):
		self.debugLog("Starting runConcurrentThread.")
		try:
			while True:
				# Check for newer plugin versions.
				self.updater.checkVersionPoll()
				
				# Clear the alternate weather XML content so it can be
				#   updated if needed.
				self.alternateWeatherXml = ""
				self.alternateWeatherMapXml = ""
				# Sleep (15 minutes) before updating.
				self.sleep(15 * 60)
				# Cycle through each station to update it.
				for deviceId in self.deviceList:
					# Call the update method with the device instance
					self.update(indigo.devices[deviceId])
		except self.StopThread:
			pass
	
	########################################
	def getAlternateData(self, device, stationLat, stationLong):
		# Get data from an alternate NOAA source.
		# Round the latitude and longitude data.
		stationLat = round(stationLat, 2)
		stationLong = round(stationLong, 2)
		self.debugLog("getAlternateData called. Looking for a station at coordinates %s, %s." % (stationLat, stationLong))
		
		theData = dict()
		l = None
		a = None
		listDocTree = None
		error = False
		# Indicate whether the station list was found.
		weatherStationFound = False
		stationUrl = ""
		
		# If alternate weather data hasn't yet been downloaded, do so now.
		if len(self.alternateWeatherXml) == 0:
			
			# If the station map XML has already been downloaded recently,
			#   don't download it again as it's a relatively large file.
			if len(self.alternateWeatherMapXml) == 0:
				# Download the weather station list XML file.
				self.debugLog("Device " + device.name + " downloading weather station list from " + weatherStationListUrl)
				try:
					l = urllib2.urlopen(weatherStationListUrl)
					self.debugLog("Device " + device.name + " reading returned weather station list file.")
					self.alternateWeatherMapXml = l.read()
					# Close the file.
					self.debugLog("Closing station list file.")
					l.close()
				except urllib2.HTTPError, e:
					indigo.server.log(u"Warning: HTTP error getting weather station list.", device.name)
					error = True
				except Exception, e:
					indigo.server.log(u"Unknown error getting weather station list.", device.name)
					error = True
			else:
				self.debugLog("Device " + device.name + " weather station list already downloaded. Using existing list.")
			
			self.debugLog("Device " + device.name + " parsing weather station list XML data.")
			try:
				listDocTree = parseString(self.alternateWeatherMapXml)
			except Exception, e:
				indigo.server.log(u"Warning: Error parsing weather station list XML from NOAA: %s" % (str(e)), device.name)
				error = True
			
			# Get the weather station elements from the document tree.
			if not error:
				try:
					self.debugLog("Searching for stations with the same latitude and longitude.")
					stations = listDocTree.getElementsByTagName('station')
					for station in stations:
						theLat = round(float(station.getElementsByTagName('latitude')[0].childNodes[0].data), 2)
						theLong = round(float(station.getElementsByTagName('longitude')[0].childNodes[0].data), 2)
						if theLat == stationLat and theLong == stationLong:
							self.debugLog("Found matching station.")
							stationUrl = station.getElementsByTagName('xml_url')[0].childNodes[0].data
							self.debugLog("Station observation data URL: %s" % (stationUrl))
							break		# No need to search further.
				except IndexError, e:
					indigo.server.log(u"Warning: Unable to find any \"station\" XML elements in the weather station list returned from NOAA.", device.name)
					error = True
				except Exception, e:
					indigo.server.log(u"Warning: Error obtaining alternate NOAA weather source information. %s" % (str(e)), device.name)
					error = True
			
				if len(stationUrl) > 0:
					# Get the weather information from the alternate source.
					#   Code for this process is a modified version of code
					#   taken from the "NOAA Weather" plugin created by
					#   Perceptive Automation who had no involvement in the
					#   creation of this plugin.
					try:
						a = urllib2.urlopen(stationUrl)
					except urllib2.HTTPError, e:
						self.errorLog("HTTP error getting weather information from alternate NOAA source: %s" % (str(e)))
						error = True
					except Exception, e:
						error = True
						if "nodename nor servname provided" in str(e):
							# DNS/Network error:
							self.errorLog("Unable to find the NOAA server. Possible network connection problem. Ensure the network connection on this host is working properly.")
						else:
							# Other error:
							self.errorLog("Unknown error getting weather information from alternate NOAA source: %s" % (str(e)))
					if error == False:
						self.alternateWeatherXml = a.read()
						theDocTree = parseString(self.alternateWeatherXml)
						
						# Parse the alternate weather source elements.
						try:
							observationDate = theDocTree.getElementsByTagName('observation_time_rfc822')[0].childNodes[0].data
						except IndexError, e:
							indigo.server.log(u"Warning: Alternate weather data source didn't contain an observation date. Possible corrupt data file from NOAA.", device.name)
							theData['observationDate'] = ""
						
						for state, fieldName in alternateDataFields.iteritems():
							try:
								newValue = theDocTree.getElementsByTagName(fieldName)[0].childNodes[0].data
							except IndexError, e:
								newValue = ""
							if newValue == "":	# if it's blank.
								newValue = ""
							theData[state] = newValue
						
						# Close the file.
						a.close()
						
		else:
			# Alternate data is already downloaded. Use that.
			self.debugLog("Device " + device.name + " alternate weather data already downloaded. Using that.")
			
			theDocTree = parseString(self.alternateWeatherXml)
			
			# Parse the alternate weather source elements.
			try:
				observationDate = theDocTree.getElementsByTagName('observation_time_rfc822')[0].childNodes[0].data
			except IndexError, e:
				indigo.server.log(u"Warning: Alternate weather data source didn't contain an observation date. Possible corrupt data file from NOAA.", device.name)
				theData['observationDate'] = ""
			
			for state, fieldName in alternateDataFields.iteritems():
				try:
					newValue = theDocTree.getElementsByTagName(fieldName)[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				if newValue == "":	# if it's blank.
					newValue = ""
				theData[state] = newValue
		
		self.debugLog("getAlternateData: Data returned: %s" % (str(theData)))
		
		# Indicate if there were errors in the returned data.
		if error == True:
			theData['error'] = error
		
		# Return the dictionary containing the alternate data content.
		return theData
	
	########################################
	def update(self, device):
		self.debugLog("Updating weather for device: " + device.name)
		
		# Alternate NOAA data dictionary.
		alternateData = dict()
		
		stationLat = None
		stationLong = None
		theDocTree = None
		theElements = None
		theCurrentObs = None
		theForecast = None
		newValue = ""
		tryCount = 1
		error = False
		
		# Define the URL and download the file
		theUrl = theUrlBase + u"lat=" + device.pluginProps["lat"] + u"&lon=" + device.pluginProps["long"]
		self.debugLog("Device " + device.name + " using URL " + theUrl)
		# Try multiple times to get data if parsing it fails.
		while tryCount < 5:
			self.debugLog("Attempt # " + str(tryCount) + " at downloading and parsing weather data.")
			try:
				f = urllib2.urlopen(theUrl)
			except urllib2.HTTPError, e:
				self.errorLog("HTTP error getting weather for coordinates %s, %s: %s" % (device.pluginProps['lat'], device.pluginProps['long'], str(e)))
				error = True
			except Exception, e:
				self.errorLog("Unknown error getting weather for coordinates %s, %s: %s" % (device.pluginProps['lat'], device.pluginProps['long'], str(e)))
				error = True
			
			# If there's an error obtaining the data feed, all we can do is try again until
			#   all the retry attempts are exhausted.
			if error:
				if tryCount < 5:
					# Start the loop over again.
					tryCount += 1
					continue
				else:
					return
			
			# Get the content returned from the URLs.
			self.debugLog("Device " + device.name + " reading returned weather data.")
			try:
				theXml = f.read()
			except Exception, e:
				self.errorLog(u"Error reading data from the URL " + theUrl + u". " + str(e))
			
			# Check to see if NOAA reports no forecast for the specified location.
			if "A point forecast is unavailable for the requested location" in theXml or "Invalid zone code" in theXml:
				self.errorLog("NOAA does not have weather information for the location %s, %s." % (device.pluginProps['lat'],device.pluginProps['long']))
				error = True
			
			# Convert the content to an XML Document Object Model tree.
			self.debugLog("Device " + device.name + " parsing weather XML data.")
			try:
				theDocTree = parseString(theXml)
			except Exception, e:
				self.errorLog("Error parsing XML from NOAA for device %s: %s. Try using GPS coordinaetes for a nearby airport (NOAA often places weather stations at airports)." % (device.name, str(e)))
				error = True
			# If there were errors, increment the counter. If not, set it to the max to exit the loop.
			if error:
				tryCount += 1
				error = False
			else:
				tryCount = 5
			# Close the file.
			f.close()
		
		# If there is still an error after tryCount reaches 5, we must abort.
		if error:
			return
		
		# Assign the current observation and the forecast to different variables.
		self.debugLog("Device " + device.name + " locating \"data\" elements in XML.")
		try:
			theElements = theDocTree.getElementsByTagName('data')
		except IndexError, e:
			self.errorLog("Device " + device.name + ": Unable to find any \"data\" XML elements in the returned data from NOAA.")
			return
		except AttributeError:
			self.errorLog("Unable to find current condition or forecast data in the resulting data file. Try using different GPS coordinates.")
			return
		except Exception, e:
			self.errorLog("Error parsing XML from NOAA for device " + device.name + ": " + str(e) + ". Try using GPS coordinaetes for a nearby airport (NOAA often places weather stations at airports).")
			return
		
		for tempElement in theElements:
			if tempElement.getAttribute('type') == "forecast":
				self.debugLog("Device " + device.name + " found forecast data.")
				theForecast = tempElement
			elif tempElement.getAttribute('type') == "current observations":
				self.debugLog("Device " + device.name + " found current observation data.")
				theCurrentObs = tempElement
		
		#
		# CURRENT CONDITIONS…
		#
		
		# First try to get the observation source coordinates
		#   in case we need to load the alternate NOAA data source.
		self.debugLog("Device " + device.name + " looking for observation location coordinate data.")
		try:
			# Get the observation source coordinates.
			stationLat = float(theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('latitude'))
			stationLong = float(theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('longitude'))
			self.debugLog("Device " + device.name + " found latitude %s, longitude %s." % (str(stationLat), str(stationLong)))
		
		except IndexError, e:
			indigo.server.log(u"Warning: Unable to determine appropriate source for alternate NOAA data. %s" % (str(e)), device.name)
		except Exception, e:
			indigo.server.log(u"Warning: Error while attempting to determine source for alternate NOAA data. %s" % (str(e)), device.name)
		
		# Get the observation time.  If it's not in the XML, attempt to get the alternate data.
		self.debugLog("Device " + device.name + " looking for observation time.")
		observationTime = ""	# Define the variable first.
		try:
			observationTime = theCurrentObs.getElementsByTagName('time-layout')[0].getElementsByTagName('start-valid-time')[0].childNodes[0].data
			self.debugLog("Device " + device.name + " observation time listed is \"%s.\"" % (observationTime))
		except Exception, e:
			self.debugLog("Exception encountered trying to obtain observation time: " + str(e))
			indigo.server.log(u"Observation time is unavailable or blank. Obtaining data from alternate source.", device.name)
			# Load the alternate NOAA data
			if stationLat != None and stationLong != None:
				alternateData = self.getAlternateData(device, stationLat, stationLong)
			
			if alternateData.get('observationTime', False):
				self.debugLog("Using alternate source for observation time.")
				observationTime = alternateData['observationTime']
		
		# If the observationTime is still unknown, set it to blank.
		if observationTime == "" or observationTime == "NA":
			observationTime = "T"
		
		# Convert the observation date/time to a more standardized date/time string.
		# self.debugLog("Device " + device.name + " formatting the observation time.")
		theDate, theTime = observationTime.split('T')
		theTime = theTime.split('-')[0]
		observationTime = theDate + " " + theTime
		self.debugLog("Device " + device.name + " observation time (after formatting): %s." % (observationTime))
		
		# If the observation time is blank (which can happens if the observation point is
		#   near or on a large body of water), set the observation time to "".
		if observationTime == ' ':
			observationTime = ""
			self.debugLog("Device " + device.name + " observastion time blank. Setting to %s." % (observationTime))
		
		device.updateStateOnServer(key="observationTime", value=observationTime)
		
		# Get the more information link and assign it.
		try:
			state = "weatherHistoryURL"
			newValue = theCurrentObs.getElementsByTagName('moreWeatherInformation')[0].childNodes[0].data
		except IndexError, e:
			newValue = ""
		except Exception, e:
			indigo.server.log(u"Warning: Error obtaining the more weather info URL from the data feed. %s" % (str(e)), device.name)
		self.updateDeviceState(device, state, newValue)
		
		# Go through each status field and try to assign the right value.
		for state in currentConditions:
			newValue = ''	# Clear this commonly used variable fist.
			if state == "currentCondition":
				#
				# Current Condition
				self.debugLog("Device " + device.name + " current condition element found.")
				try:
					if theCurrentObs != None:
						newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('weather')[0].getElementsByTagName('weather-conditions')[0].getAttribute('weather-summary')
						self.debugLog("Device " + device.name + " current condition found is %s." % (newValue))
					else:
						# Setting newValue to ""
						newValue = ""
				except IndexError, e:
					newValue = ""
					self.debugLog("Device " + device.name + " current condition not found. Setting to %s." % (newValue))
				if newValue == "":
					newValue = ""
					self.debugLog("Device " + device.name + " current condition is blank. Seting to %s." % (newValue))
				# If this data isn't available, try to get it from the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " current condition is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for current condtion information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			
			elif state == "currentConditionIcon":
				#
				# Current Condition Icon
				self.debugLog("Device " + device.name + " current condition icon element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('conditions-icon')[0].getElementsByTagName('icon-link')[0].childNodes[0].data
					self.debugLog("current condition icon URL: " + newValue)
					#newValue = newValue.replace("http://forecast.weather.gov/images/wtf/", "")
					# 02-Jul-2012. It seems NOAA has started putting some icons in a subfolder. Removing it's reference.
					#newValue = newValue.replace("medium/", "")
					# 24-Jul-2015. It seems NOAA has once again changed the base URL of weather condition icons.
					#    It's time to extract the icon filename correctly with a Regular Expression.
					newValue = re.sub("^.*/", "", newValue)
					self.debugLog("current condition icon now: " + newValue)
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("current condition icon could not be found. Error was: " + str(e))
				if newValue == "NULL":	# NWS uses "NULL" to indicate an icon isn't available.
					newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " current condition icon is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for current condtion icon information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "temperatureF":
				#
				# Current Temperature (F)
				self.debugLog("Device " + device.name + " current temperature (F) element found.")
				try:
					# Sometiems the data can contain more than 1 temperature element.
					#   Find the correct one.
					tempElements = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')
					# Loop through the tempElements to find the max and min temperature.
					for a in tempElements:
						if a.getAttribute('type') == "apparent":
							self.debugLog("Device " + device.name + " current temperature (F) element type is \"apparent\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
						elif a.getAttribute('type') == "air":
							self.debugLog("Device " + device.name + " current temperature (F) element type is \"air\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("current temperature (F) could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " temperature (F) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for temperature (F) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "temperatureC":
				#
				# Current Temperature (C)
				self.debugLog("Device " + device.name + " current temperature (C) element found.")
				try:
					# Sometiems the data can contain more than 1 temperature element.
					#   Find the correct one.
					tempElements = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')
					# Loop through the tempElements to find the max and min temperature.
					for a in tempElements:
						if a.getAttribute('type') == "apparent":
							self.debugLog("Device " + device.name + " current temperature element type is \"apparent\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
						elif a.getAttribute('type') == "air":
							self.debugLog("Device " + device.name + " current temperature element type is \"air\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("current temperature (C) could not be found. Error was: " + str(e))
				if newValue != "":
					try:
						newValue = round(((float(newValue)-32)*5/9),0)
					except ValueError:
						newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " current temperature (C) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for temperature (C) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "humidity":
				#
				# Relative Humidity (percent)
				self.debugLog("Device " + device.name + " relative humidity element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('humidity')[0].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("relative humidity could not be found. Error was: " + str(e))
				# Sometimes NWS includes the percent character in the results.  Remove it.
				newValue = newValue.rstrip("%")
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " humidity is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for humidity information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "dewPointF":
				#
				# Dew Point (F)
				self.debugLog("Device " + device.name + " dew point (F) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[1].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("dew point (F) could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " dew point (F) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for dew point (F) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "dewPointC":
				#
				# Dew Point (C)
				self.debugLog("Device " + device.name + " dew point (C) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[1].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("dew point (C) could not be found. Error was: " + str(e))
				if newValue != "":
					try:
						newValue = round(((float(newValue)-32)*5/9),0)
					except ValueError:
						newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " dew point (C) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for dew point (C) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "heatIndexF":
				#
				# Heat Index (F)
				self.debugLog("Device " + device.name + " heat index (F) element found.")
				try:
					tempTemp = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[0].getElementsByTagName('value')[0].childNodes[0].data
					tempHumid = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('humidity')[0].getElementsByTagName('value')[0].childNodes[0].data
					# Calculate the heat index from temperature and relative humidity.
					try:
						tempTemp = float(tempTemp)
						tempHumid = float(tempHumid)
						# Temperatures below 70 degrees F are meaningless.
						if tempTemp >= 70:
							# Using the formula found at http://en.wikipedia.org/wiki/Heat_index#Formula
							c1,c2,c3,c4,c5,c6,c7,c8,c9 = -42.38,2.049,10.14,-0.2248,-0.006838,-0.05482,0.001228,0.0008528,-0.00000199
							newValue = int(round((c1+c2*tempTemp+c3*tempHumid+c4*tempTemp*tempHumid+c5*tempTemp*tempTemp+c6*tempHumid*tempHumid+c7*tempTemp*tempTemp*tempHumid+c8*tempTemp*tempHumid*tempHumid+c9*tempTemp*tempTemp*tempHumid*tempHumid),0))
						else:
							newValue = ""
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("heat index (F) could not be found. Error was: " + str(e))
				tempTemp,tempHumid = 0,0	# Clear the variables.
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " heat index (F) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for heat index (F) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "heatIndexC":
				#
				# Heat Index (C)
				self.debugLog("Device " + device.name + " heat index (C) element found.")
				try:
					tempTemp = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[0].getElementsByTagName('value')[0].childNodes[0].data
					tempHumid = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('humidity')[0].getElementsByTagName('value')[0].childNodes[0].data
					# Calculate the heat index from temperature and relative humidity.
					try:
						tempTemp = float(tempTemp)
						tempHumid = float(tempHumid)
						# Temperatures below 70 degrees F are meaningless.
						if tempTemp >= 70:
							# Using the formula found at http://en.wikipedia.org/wiki/Heat_index#Formula
							c1,c2,c3,c4,c5,c6,c7,c8,c9 = -42.38,2.049,10.14,-0.2248,-0.006838,-0.05482,0.001228,0.0008528,-0.00000199
							newValue = (c1+c2*tempTemp+c3*tempHumid+c4*tempTemp*tempHumid+c5*tempTemp*tempTemp+c6*tempHumid*tempHumid+c7*tempTemp*tempTemp*tempHumid+c8*tempTemp*tempHumid*tempHumid+c9*tempTemp*tempTemp*tempHumid*tempHumid)
							newValue = int(round(((newValue-32)*5/9),0))
						else:
							newValue = ""
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("heat index (C) could not be found. Error was: " + str(e))
				tempTemp,tempHumid = 0,0	# Clear the variables.
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " heat index (C) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for heat index (C) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "waterTemperatureF":
				#
				# Current Water Temperature (F)
				self.debugLog("Device " + device.name + " current water temperature (F) element found.")
				try:
					# Sometiems the data can contain more than 1 temperature element.
					#   Find the correct one.
					tempElements = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')
					# Loop through the tempElements to find the max and min temperature.
					for a in tempElements:
						if a.getAttribute('type') == "water":
							self.debugLog("Device " + device.name + " current water temperature (F) element type is \"apparent\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("water temperature (F) could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " temperature (F) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for temperature (F) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "waterTemperatureC":
				#
				# Current Water Temperature (C)
				self.debugLog("Device " + device.name + " current water temperature (C) element found.")
				try:
					# Sometiems the data can contain more than 1 temperature element.
					#   Find the correct one.
					tempElements = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')
					# Loop through the tempElements to find the max and min temperature.
					for a in tempElements:
						if a.getAttribute('type') == "water":
							self.debugLog("Device " + device.name + " current water temperature (C) element type is \"water\".")
							newValue = a.getElementsByTagName('value')[0].childNodes[0].data
							break
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("water temperature (C) could not be found. Error was: " + str(e))
				if newValue != "":
					try:
						newValue = round(((float(newValue)-32)*5/9),0)
					except ValueError:
						newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " current temperature (C) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for temperature (C) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "visibility":
				#
				# Visibility (miles)
				self.debugLog("Device " + device.name + " visibility element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('weather')[0].getElementsByTagName('weather-conditions')[1].getElementsByTagName('value')[0].getElementsByTagName('visibility')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("visibility could not be found. Error was: " + str(e))
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				# If this data isn't available, try to get it from the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " visibility is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for visibility information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "pressureInches":
				#
				# Barometric Pressure (Inches Hg)
				self.debugLog("Device " + device.name + " pressure (inches Hg) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('pressure')[0].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("pressure (inches Hg) could not be found. Error was: " + str(e))
				if newValue == "NA" or "mb" in newValue:
					# NWS uses "NA" to indicate not available.  In some cases (usually
					#   off-shore weather buoys), the millibars measurement is incorrectly
					#   placed into the inches value.  Check for this.
					newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " pressure (inches) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for pressure (inches) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "pressureMillibars":
				#
				# Barometric Pressure (Millibars)
				self.debugLog("Device " + device.name + " pressure (millibars) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('pressure')[0].getElementsByTagName('value')[0].childNodes[0].data
					# Convert the Inches of Hg to Millibars, then round to a whole number.
					try:
						newValue = int(round((float(newValue) * 33.863886666),0))
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("pressure (millibars) could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " pressure (millibars) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for pressure (millibars) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "windDegrees":
				#
				# Wind Degrees (degrees)
				self.debugLog("Device " + device.name + " wind direction degrees element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('direction')[0].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("wind direction degrees could not be found. Error was: " + str(e))
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				if newValue != "":
					# If the data is not a number (as it should be), set it to blank.
					try:
						int(newValue)
					except ValueError:
						newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " wind direction degrees are unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for wind direction degrees information.")
						newValue = alternateData[state]
				# Check for negative numbers (sometimes NWS uses this to indicate variable direction).
				if newValue != '':
					if int(newValue) < 0:
						newValue = 999
				else:	# Sometimes the value is blank.
					newValue = 999
				
				self.updateDeviceState(device, state, newValue)
			elif state == "windDirection":
				#
				# Wind Direction
				self.debugLog("Device " + device.name + " wind direction element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('direction')[0].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("wind direction could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " wind direction is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for wind direction information.")
						newValue = alternateData['windDegrees']
				
				# Convert the compass degree number to a cardinal name.
				try:
					tempValue = int(newValue)
					if tempValue > 360 or tempValue < 0:
						# NWS uses -10 and 999 to designate variable wind direction.
						newValue = "Variable"
					elif tempValue < 12 or tempValue > 348:
						newValue = "North"
					elif tempValue < 34:
						newValue = "North-Northeast"
					elif tempValue < 57:
						newValue = "Northeast"
					elif tempValue < 79:
						newValue = "East-Northeast"
					elif tempValue < 102:
						newValue = "East"
					elif tempValue < 124:
						newValue = "East-Southeast"
					elif tempValue < 147:
						newValue = "Southeast"
					elif tempValue < 169:
						newValue = "South-Southeast"
					elif tempValue < 192:
						newValue = "South"
					elif tempValue < 214:
						newValue = "South-Southwest"
					elif tempValue < 237:
						newValue = "Southwest"
					elif tempValue < 259:
						newValue = "West-Southwest"
					elif tempValue < 282:
						newValue = "West"
					elif tempValue < 304:
						newValue = "West-Northwest"
					elif tempValue < 327:
						newValue = "Northwest"
					elif tempValue < 349:
						newValue = "North-Northwest"
				except ValueError:
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "windKnots":
				#
				# Wind Speed (knots)
				self.debugLog("Device " + device.name + " wind speed (knots) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[1].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("wind speed (knots) could not be found. Error was: " + str(e))
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				# Sometimes NWS will include non-numbers in the result. Remove the non-number characters
				#   and log the error.
				if re.match("\D", newValue) != None:
					indigo.server.log("Warning: Wind speed data contains non-numeric characters. Data may be corrupt or inaccurate.")
					newValue = re.sub("\D", "", newValue)
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " wind speed (knots) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for wind speed (knots) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "windMPH":
				#
				# Wind Speed (MPH)
				self.debugLog("Device " + device.name + " wind speed (MPH) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[1].getElementsByTagName('value')[0].childNodes[0].data
					# Convert knots to MPH (MPH = Knots * 1.15155).
					# We round to nearest integer.
					try:
						newValue = int(round((float(newValue) * 1.15155),0))
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("wind speed (MPH) could not be found. Error was: " + str(e))
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " wind speed (MPH) is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for wind speed (MPH) information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "windGustKnots":
				#
				# Wind Gust Speed (knots)
				self.debugLog("Device " + device.name + " gust wind speed (knots) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[0].getElementsByTagName('value')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("gust wind speed (knots) could not be found. Error was: " + str(e))
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "windGustMPH":
				#
				# Wind Gust Speed (MPH)
				self.debugLog("Device " + device.name + " gust wind speed (MPH) element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[0].getElementsByTagName('value')[0].childNodes[0].data
					# Convert knots to MPH (MPH = Knots * 1.15155).
					# We round to nearest integer.
					try:
						newValue = int(round((float(newValue) * 1.15155),0))
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("gust wind speed (MPH) could not be found. Error was: " + str(e))
				self.updateDeviceState(device, state, newValue)
			elif state == "windChillF":
				#
				# Wind Chill Temperature (degrees F)
				self.debugLog("Device " + device.name + " wind chill (F) element found.")
				try:
					tempSpeed = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[0].getElementsByTagName('value')[0].childNodes[0].data
					# If the data isn't available, try to get the alternate source.
					if tempSpeed == "NA":
						# Load the alternate NOAA data
						self.debugLog("Device " + device.name + " wind speed for wind chill is unavailable. Checking alternate data source.")
						if stationLat != None and stationLong != None:
							alternateData = self.getAlternateData(device, stationLat, stationLong)
						
						if alternateData.get('windMPH', False):
							self.debugLog("Using alternate source for wind speed information.")
							tempSpeed = alternateData['windMPH']
						
					self.debugLog("Device " + device.name + " wind chill (F): using " + str(tempSpeed) + " as wind speed.")
					tempTemp = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[0].getElementsByTagName('value')[0].childNodes[0].data
					self.debugLog("Device " + device.name + " wind chill (F): using " + str(tempTemp) + " as temperature.")
					# Calculate wind chill using the formula found
					#   at http://www.nws.noaa.gov/om/windchill/
					try:
						tempSpeed = float(tempSpeed)
						tempTemp = float(tempTemp)
						# Wind chill temperatures are only valid at 50 degrees F or below.
						if tempTemp <= 50:
							newValue = (35.74+0.6215*tempTemp-35.75*(tempSpeed**0.16)+0.4275*tempTemp*(tempSpeed**0.16))
							# Round to a whole number.
							newValue = int(round(newValue,0))
						else:
							newValue = ""
					except ValueError, e:
						self.debugLog("Value error. Wind speed could not be calculated: " + str(e))
						newValue = ""
					except Exception, e:
						self.debugLog("Unable to calculate wind speed: " + str(e))
				except IndexError, e:
					self.debugLog("Element does not exist for wind speed calculation: " + str(e))
					newValue = ""
				except Exception, e:
					self.debugLog("wind chill (F) could not be found. Error was: " + str(e))
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "windChillC":
				#
				# Wind Chill Temperature (degrees C)
				self.debugLog("Device " + device.name + " wind chill (C) element found.")
				try:
					tempSpeed = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('wind-speed')[0].getElementsByTagName('value')[0].childNodes[0].data
					# If the data isn't available, try to get the alternate source.
					if tempSpeed == "NA":
						# Load the alternate NOAA data
						self.debugLog("Device " + device.name + " wind speed for wind chill is unavailable. Checking alternate data source.")
						if stationLat != None and stationLong != None:
							alternateData = self.getAlternateData(device, stationLat, stationLong)
						
						if alternateData.get('windMPH', False):
							self.debugLog("Using alternate source for wind speed information.")
							tempSpeed = alternateData['windMPH']
						
					tempTemp = theCurrentObs.getElementsByTagName('parameters')[0].getElementsByTagName('temperature')[0].getElementsByTagName('value')[0].childNodes[0].data
					# Calculate wind chill using the formula found
					#   at http://www.nws.noaa.gov/om/windchill/
					try:
						tempSpeed = float(tempSpeed)
						tempTemp = float(tempTemp)
						# Wind chill temperatures are only valid at 50 degrees F or below.
						if tempTemp <= 50:
							newValue = (35.74+0.6215*tempTemp-35.75*(tempSpeed**0.16)+0.4275*tempTemp*(tempSpeed**0.16))
							# Convert to Celsius and round to a whole number.
							newValue = int(round(((newValue-32)*5/9),0))
						else:
							newValue = ""
					except ValueError:
						newValue = ""
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("wind chill (C) could not be found. Error was: " + str(e))
				self.updateDeviceState(device, state, newValue)
			elif state == "location":
				#
				# Location
				self.debugLog("Device " + device.name + " location element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('area-description')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("location could not be found. Error was: " + str(e))
				if newValue == "NA" or newValue.split(',')[0] == "NA":
					# NWS uses "NA" (and sometimes "NA,") to indicate not available.
					newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " location is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for location information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "latitude":
				#
				# Latitude
				self.debugLog("Device " + device.name + " latitude element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('latitude')
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("latitude could not be found. Error was: " + str(e))
				if newValue == "NA" or newValue == '':
					# NWS uses "NA" (or sometimes nothing at all) to indicate not available.
					newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " latitude is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for latitude information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "longitude":
				#
				# Longitude
				self.debugLog("Device " + device.name + " longitude element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('longitude')
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("longitude could not be found. Error was: " + str(e))
				if newValue == "NA" or newValue == '':
					# NWS uses "NA" (or sometimes nothing at all) to indicate not available.
					newValue = ""
				# If the data isn't available, try to get the alternate source.
				if newValue == "":
					# Load the alternate NOAA data
					self.debugLog("Device " + device.name + " longitude is unavailable. Checking alternate data source.")
					if stationLat != None and stationLong != None:
						alternateData = self.getAlternateData(device, stationLat, stationLong)
					
					if alternateData.get(state, False):
						self.debugLog("Using alternate source for longitude information.")
						newValue = alternateData[state]
				
				self.updateDeviceState(device, state, newValue)
			elif state == "elevation":
				#
				# Elevation
				self.debugLog("Device " + device.name + " elevation element found.")
				try:
					newValue = theCurrentObs.getElementsByTagName('location')[0].getElementsByTagName('height')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					self.debugLog("elevation could not be found. Error was: " + str(e))
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
		
		self.debugLog("Current condition processing complete.")
		
		#
		# FORECAST DATA...
		#
		self.debugLog("Beginning forecast data processing.")
		
		# These variables are used throughout this section.
		startTimes = []
		timePeriods = []
		endTimes = []
		
		# -- Get the time layouts --
		#    (referenced by most forecast parameters).
		self.debugLog("Getting forecast time layout data.")
		try:
			timeLayoutTree = theForecast.getElementsByTagName('time-layout')
		except Exception, e:
			self.debugLog("Unable to obtain forecast time layout data. Error was: " + str(e))
			timeLayoutTree = None
		
		# Get the more information link and assign it.
		self.debugLog("Getting forecast URL element.")
		try:
			state = "forecastURL"
			newValue = theForecast.getElementsByTagName('moreWeatherInformation')[0].childNodes[0].data
		except IndexError, e:
			newValue = ""
		except Exception, e:
			newValue = ""
			indigo.server.log(u"Warning: Error obtaining the forecast info URL from the data feed. %s" % (str(e)), device.name)
		self.updateDeviceState(device, state, newValue)
		
		# Get the time layout keys. These are referenced by the
		#   parameters that are in a different part of the DOM tree.
		self.debugLog("Putting the time layout into a multi-dimensional list.")
		timeLayout = []	# Entire list of time groups.
		timeGroup = []	# Individual group of start times.
		timePeriodGroup = []	# Individual group of time periods.
		endTimeGroup = []	# Individual group of end times.
		if timeLayoutTree != None:
			for a in timeLayoutTree:
				# Get the name of the time group.
				layoutKey = a.getElementsByTagName('layout-key')[0].childNodes[0].data
				# Get the times that the predictions become valid and associate them
				#   with the layout key name.
				for b in a.getElementsByTagName('start-valid-time'):
					# Add the individual start time to the group.
					timeGroup.append(b.childNodes[0].data)
					# Add the individual time period name.
					timePeriodGroup.append(b.getAttribute('period-name'))
				# Get the times that the predictions become invalid and associate them
				#   with the layout key name.
				for b in a.getElementsByTagName('end-valid-time'):
					# Add the individual end time to the group.
					endTimeGroup.append(b.childNodes[0].data)
				# Add the layout name, group of times, and group of period names to the overall layout list.
				timeLayout.append([layoutKey, timeGroup, timePeriodGroup, endTimeGroup])
				# Reset the group lists for when the loop starts over.
				timeGroup = []
				timePeriodGroup = []
				endTimeGroup = []
		
		# -- Get the forecast parameters --
		self.debugLog("Getting forecast elements.")
		forecastParams = theForecast.getElementsByTagName('parameters')[0]
		
		# Set the forecast location data.
		for state in forecastLocationData:
			if state == "forecastLocation":
				#
				# Forecast Location
				self.debugLog("Device " + device.name + " forecast location element found.")
				try:
					newValue = theForecast.getElementsByTagName('location')[0].getElementsByTagName('area-description')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					indigo.server.log(u"Warning: Error obtaining the forecast location from the data feed. %s" % (str(e)), device.name)
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "forecastLatitude":
				#
				# Forecast Latitude
				self.debugLog("Device " + device.name + " forecast latitude element found.")
				try:
					newValue = theForecast.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('latitude')
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					indigo.server.log(u"Warning: Error obtaining the forecast latitude from the data feed. %s" % (str(e)), device.name)
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "forecastLongitude":
				#
				# Forecast Longitude
				self.debugLog("Device " + device.name + " forecast longitude element found.")
				try:
					newValue = theForecast.getElementsByTagName('location')[0].getElementsByTagName('point')[0].getAttribute('longitude')
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					indigo.server.log(u"Warning: Error obtaining the forecast longitude from the data feed. %s" % (str(e)), device.name)
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
			elif state == "forecastElevation":
				#
				# Forecast Elevation
				self.debugLog("Device " + device.name + " forecast elevation element found.")
				try:
					newValue = theForecast.getElementsByTagName('location')[0].getElementsByTagName('height')[0].childNodes[0].data
				except IndexError, e:
					newValue = ""
				except Exception, e:
					newValue = ""
					indigo.server.log(u"Warning: Error obtaining the forecast elevation from the data feed. %s" % (str(e)), device.name)
				if newValue == "NA":	# NWS uses "NA" to indicate not available.
					newValue = ""
				self.updateDeviceState(device, state, newValue)
		
		#
		# CHANCE OF RAIN...
		#
		
		# Get the probability of precipitation states from the DOM.
		self.debugLog("Getting probability of precipitation data.")
		probOfPrecipElements = forecastParams.getElementsByTagName('probability-of-precipitation')[0]
		
		self.debugLog("Getting time layout for PoP data.")
		# Get the time layout.
		layoutKey = probOfPrecipElements.getAttribute('time-layout')
		self.debugLog("Time layout is \"" + layoutKey + "\".")
		
		# Loop through the time layouts to find the right one.
		for a in timeLayout:
			# We only want the time layout for this forecast parameter
			# a[0] contains the time-layout key name.
			if a[0] == layoutKey:
				self.debugLog("Found matching layout in time layout list.")
				# Found it.  Assign the list of times to a variable to use later.
				# a[2] contains the list of period names for the layout.
				startTimes = a[1]
				timePeriods = a[2]
		# Now loop through the forecastPrecipVals device states and assign those.
		self.debugLog("Looping through each PoP element.")
		# CLear existing probability of precipitation values.
		# for s in forecastPeriods:
			# s = 'probOfPrecip' + s
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'probOfPrecip' + state
			# Get the corresponding value for this state and assign it, if it exists.
			try:
				precipValueElement = probOfPrecipElements.getElementsByTagName('value')[i]
				# If the element has an "xsi:nil" attribute, the chance of rain is zero.
				if precipValueElement.getAttribute('xsi:nil') == "true":
					newValue = 0
				else:
					newValue = precipValueElement.childNodes[0].data
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the probability of precipitation from the data feed. %s" % (str(e)), device.name)
			# Probability of precipitation values appear in chronological order. However, after 3:00 PM local time,
			#   the PoP times shift to show the chance of rain for the evening first. Catch this
			#   condition by looking for evening desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('probOfPrecip' + forecastPeriods[0]):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		#
		# FORECAST CONDITIONS…
		#
		
		# Get the forecast condition elements from the DOM.
		self.debugLog("Getting forecast condition elements.")
		conditionElements = forecastParams.getElementsByTagName('weather')[0]
		#   ... as well as the icon link elements.
		self.debugLog("Getting forecast condition icon elements.")
		conditionIconElements = forecastParams.getElementsByTagName('conditions-icon')[0]
		#   ... and the worded forecasts.
		self.debugLog("Getting worded forecast elements.")
		wordedElements = forecastParams.getElementsByTagName('wordedForecast')[0]
		
		# Get the time layout.
		#   (This layout applies to conditions, icons, and worded forecasts).
		self.debugLog("Getting the time layout for weather conditions.")
		layoutKey = conditionElements.getAttribute('time-layout')
		
		# Loop through the time layouts to find the right one.
		for a in timeLayout:
			# We only want the time layout for this forecast parameter
			# a[0] contains the time-layout key name.
			if a[0] == layoutKey:
				self.debugLog("Found matching layout in time layout list.")
				# Found it.  Assign the list of times to a variable to use later.
				# a[2] contains the list of period names for the layout.
				startTimes = a[1]
				timePeriods = a[2]
		
		# Loop through the forecastConditionTimes device states and assign times from the startTimes list.
		self.debugLog("Looping through each forecast condition time element.")
		# Clear the forecastConditionTimes.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Time'
			# self.updateDeviceState(device, s, "")
		i = 0	# Used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Time'
			# Try to get the corresponding time for this state and assign it.
			try:
				# Format the time into a more standardized time.
				theDate, theTime = startTimes[i].split('T')
				theTime = theTime.split('-')[0]
				newValue = theDate + " " + theTime
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the forecast time period from the data feed. %s" % (str(e)), device.name)
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Time'):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		# Next, loop through the forecastConditionPeriod device states and assign the
		#   corresponding time period names.
		self.debugLog("Looking through forecast period name elements for the current period.")
		# Clear the forecastConditionPeriods first.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Period'
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Period'
			# Try to get the corresponding period name for this state and assign it.
			try:
				newValue = timePeriods[i]
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the forecast period from the data feed. %s" % (str(e)), device.name)
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Period'):
				if "Evening" in newValue or "Overnight" in newValue or "Tonight" in newValue:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		# Now loop through the forecastConditionTexts device states and assign the
		#   corresponding forecast conditions.
		self.debugLog("Looping through each forecast condition element.")
		# Clear the forecastConditionTexts first.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Text'
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Text'
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = conditionElements.getElementsByTagName('weather-conditions')[i].getAttribute('weather-summary')
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the forecast text from the data feed. %s" % (str(e)), device.name)
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Text'):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		# Next, loop through the forecastConditionIcons device states and assign the
		#   corresponding forecast condition icon names.
		self.debugLog("Looping through each forecast condition icon element.")
		# Clear the forecastConditionIcons first.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Icon'
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Icon'
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = conditionIconElements.getElementsByTagName('icon-link')[i].childNodes[0].data
				#newValue = newValue.replace("http://forecast.weather.gov/images/wtf/", "")
				# 02-Jul-2012. It seems NOAA has started putting some icons in a subfolder. Removing it's reference.
				#newValue = newValue.replace("medium/", "")
				# 24-Jul-2015. It seems NOAA has once again changed the base URL of weather condition icons.
				#    It's time to extract the icon filename correctly with a Regular Expression.
				newValue = re.sub("^.*/", "", newValue)
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the forecast icon from the data feed. %s" % (str(e)), device.name)
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Icon'):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		# Now, loop through the forecastsWorded device states and assign the
		#   corresponding worded forcasts.
		self.debugLog("Looping through each worded forecast element.")
		# Clear the forecastsWorded states first.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Worded'
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Worded'
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = wordedElements.getElementsByTagName('text')[i].childNodes[0].data
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the worded forecast from the data feed. %s" % (str(e)), device.name)
			if newValue == "NA":	# NWS uses "NA" to indicate not available.
				newValue = ""
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Worded'):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		# Now get the hazardous weather advisory elements.
		self.debugLog("Getting forecasted hazardouns weather advisory elements.")
		weatherAdvisories = forecastParams.getElementsByTagName('hazards')
		# Get the time layout.
		self.debugLog("Getting the time layout for hazardous weather advisory elements.")
		layout1Key = ""
		layout2Key = ""
		layout3Key = ""
		try:
			layout1Key = weatherAdvisories[0].getAttribute('time-layout')
			layout2Key = weatherAdvisories[1].getAttribute('time-layout')
			layout3Key = weatherAdvisories[2].getAttribute('time-layout')
		except Exception:
			pass
		
		# Loop through the time layouts to find the right one.
		startTimes1 = []
		startTimes2 = []
		startTimes3 = []
		timePeriods1 = []
		timePeriods2 = []
		timePeriods3 = []
		endTimes1 = []
		endTimes2 = []
		endTimes3 = []
		self.debugLog("Looking through time layout for weather advisory layouts.")
		for a in timeLayout:
			# We only want the time layout for this forecast parameter
			# a[0] contains the time-layout key name.
			if a[0] == layout1Key:
				self.debugLog("Found matching layout in time layout list for 1st weather advisory.")
				# Found it.  Assign the list of times to a variable to use later.
				# a[2] contains the list of period names for the layout.
				# a[3] contains the list of end times.
				startTimes1 = a[1]
				timePeriods1 = a[2]
				endTimes1 = a[3]
				# Format the time into a more standardized time.
				theDate, theTime = startTimes1[0].split('T')
				theTime = theTime.split('-')[0]
				startTimes1[0] = theDate + " " + theTime
				theDate, theTime = endTimes1[0].split('T')
				theTime = theTime.split('-')[0]
				endTimes1[0] = theDate + " " + theTime
			if a[0] == layout2Key:
				self.debugLog("Found matching layout in time layout list for 2nd weather advisory.")
				startTimes2 = a[1]
				timePeriods2 = a[2]
				endTimes2 = a[3]
				# Format the time into a more standardized time.
				theDate, theTime = startTimes2[0].split('T')
				theTime = theTime.split('-')[0]
				startTimes2[0] = theDate + " " + theTime
				theDate, theTime = endTimes2[0].split('T')
				theTime = theTime.split('-')[0]
				endTimes2[0] = theDate + " " + theTime
			if a[0] == layout3Key:
				self.debugLog("Found matching layout in time layout list for 3rd weather advisory.")
				startTimes3 = a[1]
				timePeriods3 = a[2]
				endTimes3 = a[3]
				# Format the time into a more standardized time.
				theDate, theTime = startTimes3[0].split('T')
				theTime = theTime.split('-')[0]
				startTimes3[0] = theDate + " " + theTime
				theDate, theTime = endTimes3[0].split('T')
				theTime = theTime.split('-')[0]
				endTimes3[0] = theDate + " " + theTime
		
		# Next, loop through the weatherAdvisory device states and assign the
		#   corresponding hazards and start/end times (if they exist).
		self.debugLog("Looping through each hazard (weather advisory) element.")
		# Clear the weatherAdvisory first.
		for s in range(1, 4):
			for s2 in ['Text', 'StartTime', 'EndTime']:
				s2 = 'weatherAdvisory' + str(s) + s2
				self.updateDeviceState(device, s2, "")
		i = 0	# Again, used to count elements.
		for weatherAdvisory in weatherAdvisories:
			if i > 2:
				# There are more than 3 hazardous weather advisories. We only use the first 3.
				#   so let's break the loop.
				break
			
			stateText = 'weatherAdvisory' + str(i+1) + 'Text'
			stateStart = 'weatherAdvisory' + str(i+1) + 'StartTime'
			stateEnd = 'weatherAdvisory' + str(i+1) + 'EndTime'
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValueText = weatherAdvisory.getElementsByTagName('hazard-conditions')[0].getElementsByTagName('hazard')[0].getAttribute('headline')
				self.debugLog("Weather advisory " + str(i+1) + ": " + newValueText)
				if i == 0:
					newValueStart = startTimes1[0]
					newValueEnd = endTimes1[0]
				elif i == 1:
					newValueStart = startTimes2[0]
					newValueEnd = endTimes2[0]
				elif i == 2:
					newValueStart = startTimes3[0]
					newValueEnd = endTimes3[0]
				self.debugLog("Start time: " + newValueStart + ", End time: " + newValueEnd)
			except Exception, e:
				newValueText = ""
				newValueStart = ""
				newValueEnd = ""
				self.debugLog("Unable to get weather advisory. %s" % (str(e)))
				pass
			self.updateDeviceState(device, stateText, newValueText)
			self.updateDeviceState(device, stateStart, newValueStart)
			self.updateDeviceState(device, stateEnd, newValueEnd)
			i += 1	# Increment the element counter.
		
		# Now, loop through the forecastsWorded device states and assign the
		#   corresponding worded forcasts.
		self.debugLog("Looping through each worded forecast element.")
		# Clear the forecastsWorded states first.
		# for s in forecastPeriods:
			# s = 'forecast' + s + 'Worded'
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in forecastPeriods:
			state = 'forecast' + state + 'Worded'
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = wordedElements.getElementsByTagName('text')[i].childNodes[0].data
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the worded forecast from the data feed. %s" % (str(e)), device.name)
			if newValue == "NA":	# NWS uses "NA" to indicate not available.
				newValue = ""
			# Forecast values appear in chronological order. However, after 3:00 PM local time,
			#   the forecast times shift to show the forecast for the evening. Catch this
			#   condition by looking for evening name desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == ('forecast' + forecastPeriods[0] + 'Worded'):
				if "Evening" in timePeriods[i] or "Overnight" in timePeriods[i] or "Tonight" in timePeriods[i]:
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		#
		# FORECAST TEMPERATURES
		#
		
		# Get the temperature elements from the DOM.
		self.debugLog("Getting forecast temperature elements.")
		tempElements = forecastParams.getElementsByTagName('temperature')
		
		# Loop through the tempElements to find the max and min temperature.
		for a in tempElements:
			if a.getAttribute('type') == "maximum":
				self.debugLog("Found high temperatures.")
				maxTempElements = a
			elif a.getAttribute('type') == "minimum":
				self.debugLog("Found low temperatures.")
				minTempElements = a
		
		#
		# High Temperatures…
		#
		
		# Get the time layout.
		self.debugLog("Getting high temperature time loyout.")
		layoutKey = maxTempElements.getAttribute('time-layout')
		
		# Loop through the time layouts to find the right one.
		self.debugLog("Looking for matching time layout in multi-dimensional list.")
		for a in timeLayout:
			# We only want the time layout for this forecast parameter
			# a[0] contains the time-layout key name.
			if a[0] == layoutKey:
				self.debugLog("Found matching time layout " + layoutKey)
				# Found it.  Assign the list of times to a variable to use later.
				# a[1] contains the list of times for the layout specified in a[0].
				# a[2] contains the list of period names for the layout.
				startTimes = a[1]
				timePeriods = a[2]
		
		# Now loop through the maxTempVals device states and assign the corresponding maximum temperatures.
		self.debugLog("Populating high temperatures and time preiods and assigning them to device stats.")
		# Clear the maxTemp values first.
		# for s in maxTempVals:
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in maxTempVals:
			# Try to get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = maxTempElements.getElementsByTagName('value')[i].childNodes[0].data
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the high temperature from the data feed. %s" % (str(e)), device.name)
			if newValue == "NA":	# NWS uses "NA" to indicate not available.
				newValue = ""
			# Assign the value to the corresponding maxTemp<day> state.
			self.debugLog("Matching high temperature element " + str(i) + " to time period from layout.")
			# Temperature values appear in chronological order. However, after 3:00 PM local time,
			#   the high temperature times shift to show the high for the next day. Catch this
			#   condition by looking for current day desciptive text in the timePeriods[i] value.
			#   If this is the first itteration of the loop and the first high temperature period
			#   isn't for today, skip the rest of the code in the loop and go to the next state name.
			if state == maxTempVals[0]:
				if "Monday" in timePeriods[i] or "Tuesday" in timePeriods[i] or "Wednesday" in timePeriods[i] or "Thursday" in timePeriods[i] or "Friday" in timePeriods[i] or "Saturday" in timePeriods[i] or "Sunday" in timePeriods[i]:
					self.debugLog("High temperature is for next day. Assigning this temperature to next maxTemp state.")
					continue
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
		
		#
		# Low Temperatures…
		#
		
		# Get the time layout.
		self.debugLog("Getting low temperature time loyout.")
		layoutKey = minTempElements.getAttribute('time-layout')
		
		# Loop through the time layouts to find the right one.
		self.debugLog("Looking for matching time layout in multi-dimensional list.")
		for a in timeLayout:
			# We only want the time layout for this forecast parameter
			# a[0] contains the time-layout key name.
			if a[0] == layoutKey:
				self.debugLog("Found matching time layout " + layoutKey)
				# Found it.  Assign the list of times to a variable to use later.
				# a[2] contains the list of period names for the layout.
				startTimes = a[1]
				timePeriods = a[2]
		
		# Now loop through the minTempVals device states and assign the corresponding minimum temperatures.
		self.debugLog("Populating low temperatures and assigning them to device stats.")
		# Clear the existing minTemp states first.
		# for s in minTempVals:
			# self.updateDeviceState(device, s, "")
		i = 0	# Again, used to count elements.
		for state in minTempVals:
			# Get the corresponding value for this state and assign it, if it exists.
			try:
				newValue = minTempElements.getElementsByTagName('value')[i].childNodes[0].data
			except IndexError, e:
				newValue = ""
			except Exception, e:
				newValue = ""
				indigo.server.log(u"Warning: Error obtaining the low temperature from the data feed. %s" % (str(e)), device.name)
			if newValue == "NA":	# NWS uses "NA" to indicate not available.
				newValue = ""
			self.updateDeviceState(device, state, newValue)
			i += 1	# Increment the element counter.
	
	########################################
	def updateDeviceState(self,device,state,newValue):
		self.debugLog("updateDeviceState called.")
		if (newValue != device.states[state]):
			# Modify the value if the state is an Icon image.
			if "Icon" in state:
				# Interpret "DualImage.php" URLs based on the first image argument in the URL.
				if "DualImage.php" in newValue:
					# Take everything after "DualImage.php?"
					newValue = newValue.split(".php?")[1]
					# Now take the first image parameter value.
					newValue = newValue.split("i=")[1]
					# Now strip any subsequent parameters.
					newValue = newValue.split("&")[0]
				newValue = newValue.split(".")[0]
			device.updateStateOnServer(key=state, value=newValue)
	
	########################################
	def getDeviceConfigUiValues(self, valuesDict, typeId, devId):
		self.debugLog("getDeviceConfigUiValues called.")
		theLat = ''
		theLong = ''
		errorsDict = indigo.Dict()
		try:
			tempVar = valuesDict['lat']
			tempVar = valuesDict['long']
		except KeyError:
			(theLat, theLong) = indigo.server.getLatitudeAndLongitude()
			theLat = str(round(theLat,4))
			theLong = str(round(theLong,4))
			(valuesDict['lat'], valuesDict['long']) = (theLat, theLong)
		except Exception, e:
			errorsDict['lat'] = u"The latitude and longitude values are not valid. Please double check the values and try again."
			errorsDict['long'] = u"The latitude and longitude values are not valid. Please double check the values and try again."
		theLat,theLong = None,None	# Clear the variables.
		return (valuesDict, errorsDict)
	
	########################################
	# UI Validate, Close, and Actions defined in Actions.xml:
	########################################
	def validateDeviceConfigUi(self, valuesDict, typeId, devId):
		self.debugLog("validateDeviceConfigUi called.")
		theLatitude = valuesDict['lat'].encode('ascii','ignore')
		theLongitude = valuesDict['long'].encode('ascii','ignore')
		errorsDict = indigo.Dict()
		errorsDict['showAlertText'] = None
		anError = False
		#   Latitude...
		try:
			tempNum = float(theLatitude)
		except ValueError, e:
			anError = True
			errorsDict['lat'] = "Latitude must be a number."
			errorsDict['showAlertText'] += errorsDict['lat'] + "\n\n"
		except Exception, e:
			anError = True
			errorsDict['lat'] = u"Latitude value is invalid. Check value and try again."
			errorsDict['showAlertText'] += errorsDict['lat'] + "\n\n"
		# Longitude…
		try:
			tempNum = float(theLongitude)
		except ValueError, e:
			anError = True
			errorsDict['long'] = "Longitude must be a number."
			errorsDict['showAlertText'] += errorsDict['long'] + "\n\n"
		except Exception, e:
			anError = True
			errorsDict['long'] = u"Longitude value is invalid. Check value and try again."
			errorsDict['showAlertText'] += errorsDict['long'] + "\n\n"
		# If there were any errors, report them.
		if anError == True:
			errorsDict['showAlertText'] = errorsDict['showAlertText'].strip()
			return (False, valuesDict, errorsDict)
		# Round the coordinates to 4 decimals, then re-encode as ASCII.
		theLatitude = round(float(theLatitude),4)
		theLatitude = str(theLatitude).encode('ascii','ignore')
		theLongitude = round(float(theLongitude),4)
		theLongitude = str(theLongitude).encode('ascii','ignore')
		# Assign the values back to the valuesDict.
		valuesDict['lat'] = theLatitude
		valuesDict['long'] = theLongitude
		valuesDict['address'] = theLatitude + ", " + theLongitude
		theUrl = theUrlBase + u"lat=" + theLatitude + u"&lon=" + theLongitude
		try:
			f = urllib2.urlopen(theUrl)
		except urllib2.HTTPError, e:
			errorsDict['label1'] = "The geographic location could not be found, or NOAA servers aren't responding."
			errorsDict['showAlertText'] = errorsDict['label1']
			return (False, valuesDict, errorsDict)
		except Exception, e:
			errorsDict['label1'] = "There was an error verifying the geographic location with NOAA. The error was \"" + str(e) + "\""
			errorsDict['showAlertText'] = errorsDict['label1']
			return (False, valuesDict, errorsDict)
		# Check to see if NOAA reports no forecast for the specified location.
		try:
			theXml = f.read()
		except Exception, e:
			self.debugLog(u"Error reading data from the URL " + theUrl + u". " + str(e))
			errorsDict['label1'] = u"Unable to gather data from NOAA for the provided geographic location. There was a connection error. Try again."
			errorsDict['showAlertText'] = errorsDict['label1']
			return (False, valuesDict, errorsDict)
			
		if "A point forecast is unavailable for the requested location" in theXml or "Invalid zone code" in theXml:
			errorsDict['label1'] = "NOAA does not have weather information available for the specified location."
			errorsDict['showAlertText'] = errorsDict['label1']
			return (False, valuesDict, errorsDict)

		return (True, valuesDict)
